// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of scatter
//
// scatter is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// scatter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include  "MZ.h"

#include <sstream>

#include <math.h>
#include <mathconst/mathconst.h>

#include <matvec/macros.h>


// use matvec macros, but simplify them

#define VEC(vec) MATVEC_VEC_AS_STRUCT(vec,x,y,z)

#define VCROSS(p, v1, v2)	\
  MATVEC_VEC_CROSSPROD( VEC(p), VEC(v1), VEC(v2) )

#define VDOT(v1, v2)		\
  MATVEC_VEC_DOTPROD( VEC(v1), VEC(v2))

#define VLEN(v) sqrt(VDOT((v),(v)))

#define VADD( v, v1, v2 )	\
  MATVEC_VEC_ADD( VEC(v), VEC(v1), VEC(v2) )

#define VADDTO( v1, v2 )	\
  MATVEC_VEC_ADD( VEC(v1), VEC(v1), VEC(v2) )

#define VSET( v1, v2 )	\
  MATVEC_VEC_ASSIGN( VEC(v1), VEC(v2 ) )

#define VSCALE( r, v, c )	\
  MATVEC_VEC_SCALE( VEC(r), VEC(v), c )

#define VNORM(v)	\
  MATVEC_VEC_NORM(VEC(v))


int
MZ::scatter( 
  bool in_plane,
  bool out_of_plane,
  int  iter_limit,
  double theta_def, double phi_def,
  double min_cum_prob, double max_cum_prob,
  const long id,
  const double energy,
  const DVector3& surface_norm,
  DVector3& direction,
  double& theta_o,
  double& phi_o
)
{
#ifdef DEBUG
    TraceFct tf( "MZ::scatter" );
#endif

  double sin_alpha  = -1.0 * VDOT( surface_norm, direction );
  double cos_alpha  = sqrt( 1.0 - sin_alpha * sin_alpha );

  // default scatter angles; convert from arcseconds to radians
  double theta      = theta_def;
  double phi        = phi_def;

  double delta_prob = max_cum_prob - min_cum_prob;

  int status = OK;


#ifdef DEBUG
  double alpha	    = asin( sin_alpha );
#endif // DEBUG

#ifdef DEBUG
  if ( dbflag( "zones" ) ) 
    tf.message( "zones:\t%u\t%s\n", id, _id.c_str( ) );
#endif // DEBUG
    
  // temporary direction vector to handle possible iterations
  DVector3 tdir;

  do
  {
    // so far, nothing bad has happened
    status = OK;

    tdir = direction;

    if ( in_plane
	 &&
	 lookUpScatterAngle( energy, sin_alpha, min_cum_prob + ranf( ) * delta_prob, true, theta ) )
      status |= IP_EXTRAPOLATED;

    if ( out_of_plane
	 &&
	 lookUpScatterAngle( energy, sin_alpha, min_cum_prob + ranf( ) * delta_prob, false, phi ) )
      status |= OP_EXTRAPOLATED;

    theta_o  = theta;
    phi_o = phi;

    double cos_theta = cos( theta );
    double sin_theta = sin( theta );
    double cos_phi   = cos( phi   );
    double sin_phi   = sin( phi   );
  
    // cos(alpha+theta)
    double cos_at = cos_alpha * cos_theta - sin_alpha * sin_theta;
    // sin(alpha+theta)
    double sin_at = sin_alpha * cos_theta + cos_alpha * sin_theta;
  
    double A = cos_at / cos_alpha;
    double B = A * sin_alpha * cos_phi - sin_at;
  
    DVector3 r_cross_n;
    VCROSS(r_cross_n, direction, surface_norm );

    DVector3 v;
    VSCALE(v, surface_norm, sin_alpha);
    VADDTO(v, direction);
    VSCALE(v, v, 1/cos_alpha );

#ifdef DEBUG
    DVector3 u;

    VCROSS( u, direction, surface_norm );
    VNORM(u);

    //DVector3 v;
    //VCROSS( v, surface_norm, u );

#endif // DEBUG

  
    /* Construct new reflected ray. note that it's a unit vector */
  
    tdir.x = B * surface_norm.x + A * (cos_phi * direction.x + sin_phi * r_cross_n.x);
    tdir.y = B * surface_norm.y + A * (cos_phi * direction.y + sin_phi * r_cross_n.y);
    tdir.z = B * surface_norm.z + A * (cos_phi * direction.z + sin_phi * r_cross_n.z);
  
    // not quite normalized due to numerics. fix that.
    VNORM(tdir);

    // has the ray been scattered into the reflecting surface?
    if ( VDOT( tdir, surface_norm ) > 0 )
      status |= SCATTERED_INTO_SURFACE;

    // has the ray been backscattered? check that the v components
    // have the same sign (i.e. they're on the same side of the
    // surface normal in the specular plane.  if they have the same
    // sign their product will be >= 0;
    if ( VDOT( tdir, v) * VDOT( direction, v ) < 0 )
      status |= BACKSCATTERED;

#ifdef DEBUG

    if ( dbflag( "ray" ) )
      tf.message( "ray:\t(%.4f %.4f %.4f) -> (%.4f %.4f %.4f)\n", 
		  direction.x, direction.y, direction.z, tdir.x, tdir.y, tdir.z );

    if ( dbflag( "norm" ) )
      tf.message( "norm:\t(%.4f %.4f %.4f)\n", surface_norm.x, surface_norm.y, surface_norm.z );

    if ( dbflag( "theta" ) ) 
    {
      double theta_out = asin( - VDOT(tdir, surface_norm ) ) - alpha;
      tf.message( "theta_in:\t%u\t%.14g\n", id, RAD2ARCSEC(theta) );
      tf.message( "theta_out:\t%u\t%.14g\n", id, RAD2ARCSEC(theta_out) );
    }

    if ( dbflag( "phi" ) ) 
    {
      DVector3 _v;
      DVector3 _u;

      VSCALE( _v, v, VDOT(tdir, v) );

      VSCALE( _u, u, VDOT(tdir, u) );

      DVector3 l;

      VADD( l, _v, _u );

      VNORM(l);

      tf.message( "phi_in:\t\t%u\t%.14g\n", id, RAD2ARCSEC(phi) );
      tf.message( "phi_out:\t%u\t%.14g\n", id, RAD2ARCSEC(acos( VDOT(l, v ) ) * ( VDOT(l,u) < 0 ? -1 : 1 )) );
    }

#endif // DEBUG


  } while( --iter_limit && status & SCATTERED_INTO_SURFACE );

  direction = tdir;

  return status;
}


double MZ::interpolate( double x, double x1, double y1, double x2, double y2 ) {

#ifdef DEBUG
  TraceFct tf( "MZ::interpolate" );
#endif // DEBUG

  double y;
  if ( x1 == x2 ) 
    y = y1;
  else {
    double f2 = ( x - x1 ) / ( x2 - x1 );
    double f1 = 1 - f2;
    y  =  f1 * y1 + f2 * y2;
  }

  return y;
}

static Exception
cfitsioException ( string msg ) {

    char message[81];

    Exception exception;

    exception.set_message( msg );

    while ( fits_read_errmsg( message ) ) {

	exception.set_message( message );

    }

    return exception;

}

int MZ::read_chunk( long totaln, long offset, long firstn, long nvalues,
		    int narrays, iteratorCol *data,  void *userPointer ) {

    vector< Column* >* pcolumns = static_cast< vector< MZ::Column* >* > ( userPointer );

    for( int i = 0 ; i < narrays ; i++ ) {

	double* fitsdata = (double*) data[i].array + 1;

	MZ::Column* pcolumn = (*pcolumns)[i];
	pcolumn->repeat = data[i].repeat;

	pcolumn->data.insert( pcolumn->data.end(), fitsdata, fitsdata + nvalues * pcolumn->repeat );
    }

    return 0;
}



void MZ::read_scatter_table( void ) throw ( Exception ) {

    fitsfile* fptr;
    int status = 0;

    if ( fits_open_table( &fptr, _file.c_str(), READONLY, &status ) )
	throw( cfitsioException( string( "error opening scattering distribution table '")
			  + _file + "'" ) );

    if ( fits_get_num_rows( fptr, &_nrows, &status ) )
	throw( cfitsioException( string( "error getting number of rows from '" + _file + '"' ) ) );

    vector< iteratorCol > workcols(colmaps.size());
    vector< iteratorCol >::iterator workcol = workcols.begin();

    // need a vector of Column objects in the order passed to the cfitsio iterator
    vector< Column* > cfitsio_columns;

    for( ColumnMapIter colmap = colmaps.begin() ; colmap != colmaps.end() ; colmap++, workcol++ ) {

	Column* column = colmap->second;

	workcol->fptr = fptr;

	if ( fits_get_colnum( fptr, CASEINSEN, const_cast<char*>( colmap->first.c_str() ), &( workcol->colnum ), &status ) )
	    throw( cfitsioException( colmap->first + string( ": not found in '" + _file + "'" ) ) );

	workcol->datatype = TDOUBLE;
	workcol->iotype = InputCol;

	{
	    int typecode;
	    long repeat, width;
	    if ( fits_get_coltype( fptr, workcol->colnum, &typecode, &repeat, &width, &status ) )
		throw( cfitsioException( colmap->first + string( ": error getting coltype information in'" ) + _file + "'" ) );

	    column->data.reserve( _nrows * repeat);
	}

	{
	    int naxis;
	    if ( fits_read_tdim( fptr, workcol->colnum, column->naxis, &naxis, &(column->naxes[0]), &status ) )
		throw( cfitsioException( colmap->first + string( ": error getting tdim information in'") + _file + "'" ) );

	    // naxis in FITS file is for the cell vector
	    // column->naxis is for the entire column, which is nrows x naxis.
	    // however, a 1D cell vector leads to a 1D column
	    if (  !( naxis == 1 && column->naxis == 1) &&
		  ( naxis != column->naxis -1 ) ) {

		ostringstream error;
		error << colmap->first
		      << ": incorrect number of axes read. got "
		      << naxis
		      << " expected "
		      << column->naxis-1
		      << " in '" << _file << "'"
		    ;
		throw( cfitsioException( error.str() )  );

	    }

	}

	cfitsio_columns.push_back( colmap->second );
    }


    if ( fits_iterate_data( (int) workcols.size(), &(workcols[0]), 0, 0, read_chunk, &cfitsio_columns, &status ) )
	throw( cfitsioException( string("error reading data  from '" + _file + "'" ) ) );

    if ( fits_close_file( fptr, &status ) )
	throw( cfitsioException( string( "error closing scattering table '" + _file + "'" ) ) );

}


