// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of scatter
//
// scatter is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// scatter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include  "MZO.h"

#include <math.h>


MZO::MZO( const string& id, const string& filename, double zmin, double zmax ) 
    throw ( Exception ) : MZ(id, filename, zmin, zmax ) { 

  TraceFct tf("MZO::MZO"); 

  _pdist.naxis = 2;

  colmaps.insert( make_pair( "ESA",     &_esa ) );
  colmaps.insert( make_pair( "PDIST",   &_pdist ) );
  colmaps.insert( make_pair( "P0",      &_p0 ) );
  colmaps.insert( make_pair( "PTOP",    &_ptop ) );
  colmaps.insert( make_pair( "PDELTA",  &_pdelta ) );
  colmaps.insert( make_pair( "POWTOP",  &_powtop ) );
  colmaps.insert( make_pair( "NORMTOP", &_normtop ) );

  read_scatter_table();
}

bool MZO::lookUpScatterAngle( double energy, double sin_alpha, double rannum, bool inplane, double& theta ) {

#ifdef DEBUG  
  TraceFct tf( "MZO::lookUpScatterAngle" );
#endif

  double esa = energy * sin_alpha;

  double *esa_lo, *esa_hi;

  bool is_extrapolated;

  if ( NULL != bnd_bsearch( &esa, &_esa[0], _nrows, sizeof(_esa[0]),
			    (const void **) &esa_lo, (const void **)&esa_hi, esacmp ) ) {            // Exact match!!!

    is_extrapolated = false;

  } else if ( NULL == esa_lo ) { // Less than all _esa

    esa_lo  = esa_hi;
    is_extrapolated = true;

  } else if ( NULL == esa_hi ) { // Greater than all _esa

    esa_hi  = esa_lo;
    is_extrapolated = true;

  } else {                       // Within _esa

    is_extrapolated = false;
    
  }

  long idx_lo = esa_lo - &_esa[0];
  long idx_hi = esa_hi - &_esa[0];

  rannum = 2.0 * ( rannum - 0.5 );
  bool is_neg = ( rannum < 0.0 );

  if ( is_neg ) 
    rannum = -1.0 * rannum;

  double theta_esa_lo = lookUpTheta( idx_lo, rannum );
  double theta_esa_hi = lookUpTheta( idx_hi, rannum );


  if ( theta_esa_lo < 0.0 ) theta_esa_lo = 0.0;
  if ( theta_esa_hi < 0.0 ) theta_esa_hi = 0.0;

  theta = interpolate( esa, *esa_lo, theta_esa_lo, *esa_hi, theta_esa_hi );
  
  if ( is_neg )
    theta *= -1;

  if ( ! inplane )
    theta *= sin_alpha;

#ifdef DEBUG  
  if ( dbflag("theta" )) 
    tf.message( "p = %.14g theta = %.5g\n", rannum, theta );
#endif

  return is_extrapolated;
}

double MZO::lookUpTheta( const long idx, double rannum ) {

  if ( rannum <= _p0[idx] ) {

    return 0.0;

  } else if ( rannum <= _ptop[idx] ) {

    rannum = ( rannum - _p0[idx] ) / _pdelta[idx];

    long jdx = (long) rannum;

    long ndist = _pdist.naxes[0];

    if ( jdx >= ndist )
      jdx = ndist - 1;

    rannum = rannum - (double) jdx;
    double t = _pdist( idx, jdx );

    return rannum * _pdist( idx, jdx + 1) + ( 1.0 - rannum ) * t ;

  } else {

    double rtop = 0.99999; // Ouch!  this was in the old code...

    if ( rannum < rtop ) {

      return pow( _normtop[idx] * ( 1.0 - rannum ), _powtop[idx] );

    } else {

      return pow( _normtop[idx] * ( 1.0 - rtop ), _powtop[idx] );

    }

  }

}
